/*
 * Patterns for rbdrum2
 * by Javantea
 * AltSci Concepts
 * Jan 15, 2009
 *

Description:
A simple tree and a list, supported by a buf.

FIXME: This won't work properly, doh.

TODO:
partial matches.
file format
abc library

The input for the tree is quite interestingly:

abcdefgh
adefghij
blahyakd
aaaaaaaa
aaaa

The tree is created in this way:
search the root for a, add
search a for b, add
search b for c, add
search c for d, add
search d for e, add
search e for f, add
search f for g, add
search g for h, add
done
search the root for a, found
search a for d, add
search d for e, add
search e for f, add
search f for g, add
search g for h, add
done
search the root for b, add

and so forth.

This may seem idiotic, but when we compare the tree to a list, it'll make a bit 
more sense.

The way that rbdrum2 will use the patterns is this:
There is a text file with:
abcdefgh   nsalrew3
adefghij   nsalewse
blahyakd   nsalfesa
aaaaaaaa   nsalawer
aaaa       nsalgeas

The left hand side is the pattern that it will look for. It will need to be a 
tree because searching is a pita when you have 100 things to check every 20ms.
The pattern that it checks is kept as a list.
The right hand side is going to be some type of midi equivalent that I can 
output from a very simple program. It'll need to be text and hopefully human 
readable/editable since fixing sounds is a key element.
 */

#include <stdlib.h>
#include <stdio.h>

#include "pattern.h"

/**
 Do the malloc for a pattern_buf.
 */
int new_pattern_buf(pattern_buf *buf, int size)
{
	buf->buf = (unsigned char *)malloc(size);
	buf->size = 0;
	buf->pos  = 0;
	if(buf->buf != 0)
	{
		buf->size = size;
	}
	return buf->size;
}

/**
 Do the free for a pattern_buf.
 */
void pattern_buf_destroy(pattern_buf *buf)
{
	free(buf->buf);
	buf->size = 0;
	buf->pos  = 0;
}

/**
 Get a pointer that is the correct size.
 */
unsigned char *pattern_buf_get(pattern_buf *buf, int size)
{
	// Failure cases:
	if(buf->buf == 0)
		return 0;
	if(buf->pos + size > buf->size)
		return 0;
	
	unsigned char *ret = buf->buf + buf->pos;
	buf->pos += size;
	return ret;
}

/**
 Create a pattern list given a properly malloc'ed pointer.
 */
pattern_list *new_pattern_list(unsigned char *buf, int value)
{
	pattern_list *ret = 0;
	if(buf == 0)
		return 0;
	ret = (pattern_list *)buf;
	ret->next_item = 0;
	ret->value = value;
	return ret;
}

/**
 Do the free for a pattern_list (no free necessary, since buf did it).
 */
void pattern_list_destroy(pattern_list *list)
{
	if(list->next_item)
		pattern_list_destroy(list->next_item);
}

void pattern_list_add(pattern_list *list, pattern_list *item)
{
	pattern_list *current = list->next_item;
	while(current != 0)
	{
		if(current->next_item == 0)
		{
			//printf("next item %i\n", current->value);
			current->next_item = item;
			return;
		}
		//printf("next %i\n", current->value);
		current = current->next_item;
	}
	//printf("first item %i\n", list->value);
	list->next_item = item;
}

void pattern_list_pop(pattern_list *list, int index)
{
	if(index == 0)
	{
		// TODO: the reason I wrote this function is pop(0) but we can't do it.
		printf("Cannot pop 0 yet.\n");
	}
	pattern_list *current = list;
	pattern_list *prev = list;
	int i;
	for(i = 0; i < index; i++)
	{
		if(current == 0)
		{
			printf("Unable to pop %i.\n", index);
			return;
		}
		prev = current;
		current = current->next_item;
	}
	prev->next_item = current->next_item;
	// TODO: dealloc(current);
}

pattern_tree *new_pattern_tree(unsigned char *buf, int value)
{
	pattern_tree *ret = 0;
	if(buf == 0)
		return 0;
	ret = (pattern_tree *)buf;
	ret->next_sibling = 0;
	ret->first_child  = 0;
	ret->value = value;
	return ret;
}

/**
 Do the free for a pattern_tree.
 */
void pattern_tree_destroy(pattern_tree *tree)
{
	if(tree->next_sibling)
		pattern_tree_destroy(tree->next_sibling);
	if(tree->first_child)
		pattern_tree_destroy(tree->first_child);
}

void pattern_tree_add_child(pattern_tree *tree, pattern_tree *child)
{
	pattern_tree *current = tree->first_child;
	while(current != 0)
	{
		if(current->next_sibling == 0)
		{
			//printf("next sibling %i\n", current->value);
			current->next_sibling = child;
			return;
		}
		//printf("next child %i\n", current->value);
		current = current->next_sibling;
	}
	//printf("first child %i\n", tree->value);
	tree->first_child = child;
}

int pattern_tree_compare(pattern_tree *tree, pattern_list *list, pattern_tree **match)
{
	pattern_tree *found = 0;
	int found_items = 0;
	pattern_list *current_list = list;
	pattern_tree *current_tree = tree;
	while((current_list != 0) && (current_tree != 0))
	{
		//printf("compare: %i %i\n", current_tree->value, current_list->value);
		found = pattern_tree_find(current_tree, current_list->value);
		current_list = current_list->next_item;
		current_tree = found->first_child;
	}
	if(found)
	{
		// TODO: Calculate how many children exist. For now 0-3 suffice.
		found_items = 1;
		if(found->first_child)
		{
			found_items++;
			if (found->first_child->next_sibling || found->first_child->first_child)
				found_items++;
		}
	}
	*match = found;
	return found_items;
}

// Find just on this level.
pattern_tree *pattern_tree_find(pattern_tree *tree, int value)
{
	if(tree == 0)
		return 0;
	pattern_tree *ret = 0;
	pattern_tree *current = tree;
	while(current != 0)
	{
		if(current->value == value)
			return current;
		current = current->next_sibling;
	}
	return ret;
}

/**
 Add to a tree with a null-terminated string.
*/
int pattern_tree_input(pattern_tree *tree, char *tree_input, pattern_buf *buf)
{
	char *input_curr = tree_input;
	pattern_tree *prev_tree = tree;
	while(*input_curr != 0)
	{
		pattern_tree *t = pattern_tree_find(prev_tree->first_child, *input_curr);
		if(t == 0)
		{
			t = new_pattern_tree(pattern_buf_get(buf, sizeof(pattern_tree)), *input_curr);
			//printf("pattern_tree sizeof(%i), buf.pos %i\n", sizeof(pattern_tree), buf->pos);
			if(t == 0)
			{
				printf("Error with new_pattern_tree.\n");
				return 1;
			}
			pattern_tree_add_child(prev_tree, t);
		}
		else
		{
			//printf("found tree?\n");
		}
		prev_tree = t;
		input_curr++;
	}
	return 0;
}

#ifdef TEST_PATTERN

#include <stdio.h>

/**
 Test the patterns as part of pattern1.
*/
int main(int argc, char **argv)
{
	pattern_buf a;
	if(!new_pattern_buf(&a, 10240))
	{
		printf("Error with new_pattern_buf.\n");
		return 1;
	}
	
	// Create a list, length 1, data 1.
	pattern_list *l = new_pattern_list(pattern_buf_get(&a, sizeof(pattern_list)), 1);
	printf("pattern_list sizeof(%i), buf.pos %i\n", sizeof(pattern_list), a.pos);
	if(l == 0)
	{
		printf("Error with new_pattern_list.\n");
		return 1;
	}
	
	// Create a tree, length N, data blahyakdaah, ...
	char *tree_input = "blahyakdaah";
	pattern_tree *root_tree = new_pattern_tree(pattern_buf_get(&a, sizeof(pattern_tree)), 23);
	int ret = pattern_tree_input(root_tree, tree_input, &a);
	ret = pattern_tree_input(root_tree, "blahdah", &a);
	ret = pattern_tree_input(root_tree, "beee", &a);
	ret = pattern_tree_input(root_tree, "befg", &a);
	ret = pattern_tree_input(root_tree, "abcd", &a);
	
	// It's unnecessarily difficult to add a single character to our list.
	pattern_list *next = (pattern_list *)pattern_buf_get(&a, sizeof(pattern_list));
	next->value = 'b';
	pattern_list_add(l, next);
	next = (pattern_list *)pattern_buf_get(&a, sizeof(pattern_list));
	next->value = 'e';
	pattern_list_add(l, next);
	next = (pattern_list *)pattern_buf_get(&a, sizeof(pattern_list));
	next->value = 'f';
	pattern_list_add(l, next);
	next = (pattern_list *)pattern_buf_get(&a, sizeof(pattern_list));
	next->value = 'g';
	pattern_list_add(l, next);
	
	pattern_tree *found;
	int found_items = pattern_tree_compare(root_tree->first_child, l->next_item, &found);
	printf("matches: %i\n", found_items);
	
	/*
	 It seems to work thus far...
	*/
	
	pattern_list_destroy(l);
	pattern_tree_destroy(root_tree);
	
	/*
	The obvious limitation to pattern_buf is that it doesn't free when items 
	 are destroyed, so running new/destroy a bunch of times will cause this 
	 system to fail. FIXME already...
	 Lucky me, rbdrum2 only needs one list, one tree, and never needs to 
	 destroy, does it? Oh wait, it does, doh!
	Should I replace the whole damn mem thing with a ton of mallocs? Not yet.
	There's still a chance that I could do it correctly by clearing memory when 
	it suits the situation.
	
	*/
	
	pattern_buf_destroy(&a);
	return 0;
}

#endif // TEST_PATTERN
